/*
 File: mcdump.c
 Version: 1.5
 Description: PS2 Memory Card dumper
 */

/* #define PS2SDK */
#include "mcdump.h"

#define MODNAME                     "mcard_dumper"

#ifndef PS2SDK
ModuleInfo Module = {MODNAME, 0x0105};
#else
IRX_ID(MODNAME, 1, 5);
#endif

int   verbose = 1; /* default verbose level is "1" */
int (*pMcGetIoSema)() = NULL;

/* Entry point */
int Start(int argc, char *argv[])
{
 register int fd, i, j, page, port, slot, r;
 int (*pMcDetectCard)(int port, int slot);
 int (*pMcGetSpec)(int port, int slot, u_short *pagesize, u_short *blocksize, u_int *cardsize, u_char *cardflags);
 int (*pMcReadPage)(int port, int slot, int block, void *buf);
 int (*pMcGetCardType)(int port, int slot);
 char *file, *info, *mcbuffer;
 u_short pagesize, blocksize;
 u_int blocks, cardsize, erasebyte;
 u_char cardflags;
 int version, cardtype, nomlock;

 printf("IRX Memory Card Dumper V1.5\n");

 if (argc < 3)
 {
  /* not enough arguments */
  printf("Usage: mcdump -port=<port> [-slot=<slot>] [-verbose] [-silent] [-nolock] <dumpfilename> [infofilename]\n");
  return NO_RESIDENT_END;
 }

 /* setting the default parameters */ 
 file = NULL;
 info = NULL;
 port = -1;
 slot = -1;
 nomlock = 0;
 
 /* parsing module arguments */
 for (i = 1; i < argc; i ++)
 {
  if (*argv[i] != '-')
  {
   if (file == NULL) file = argv[i];
       else if (info == NULL) info = argv[i];
   continue;
  }
  
  if ((port == -1) && (!strncmp(argv[i], "-port=", 6)))
  {
   if (strlen(&argv[i][6]) > 0) port = strtol(&argv[i][6], NULL, 10);
       else printf("mcdump: Invalid -port argument.\n");
   continue;
  }

  if ((slot == -1) && (!strncmp(argv[i], "-slot=", 6)))
  {
   if (strlen(&argv[i][6]) > 0) slot = strtol(&argv[i][6], NULL, 10);
       else printf("mcdump: Invalid -slot argument.\n");
   continue;
  }
  
  if (verbose == 1)
  {
   if (!strcmp(argv[i], "-verbose")) verbose = 2;
   if (!strcmp(argv[i], "-silent")) verbose = 0;
  }
  
  if (!nomlock) if (!strcmp(argv[i], "-nolock")) nomlock = 1;
 }
 
 /* checking the parameters */
 if ((port < 0) || (port > 1))
 {
  if (port > 1) printf("mcdump: Invalid memory card port is specified.\n");
      else printf("mcdump: Memory card port is not specified\n");
  return NO_RESIDENT_END;
 }
 
 if (file == NULL)
 {
  printf("mcdump: No output file is specified.\n");
  return NO_RESIDENT_END;
 }
 
 /* using slot 0 by default unless the certain slot is specified */
 if (slot == -1) slot = 0;
 
 if (verbose)
 {
  printf("Memory Card Dumper Informations:\n");
  printf("Memory Card Port: %d\n", port);
  printf("Memory Card Slot: %d\n", slot);
  printf("Verbose Debug: %s\n", (verbose > 1) ? "ON" : "OFF");
  printf("Output Memory Card File: %s\n", file);
  printf("Output Information File: %s\n", (info != NULL) ? info : "-");
 }
 
 /* trying to use a new MCMAN module */
 version = 0x200;
 
 /* getting MCMAN's exports #21, #43 */
 VOIDPTR(pMcDetectCard) = GetLibraryEntry("mcman", version, 21);
 VOIDPTR(pMcGetSpec) = GetLibraryEntry("mcman", version, 43);

 /* getting MCMAN's export #53 */
 if (!nomlock) VOIDPTR(pMcGetIoSema) = GetLibraryEntry("mcman", version, 53);
     else pMcGetIoSema = NULL;
  
 if ((pMcDetectCard == NULL) || (pMcGetSpec == NULL))
 {
  /* trying to use an old MCMAN module (BIOS) */
  version = 0x100;
  
  /* ___McGetSpec is not available within the old MCMAN modules */
  pMcGetSpec = McGetDefSpec2;   
  /* getting address of the alternate card detection routine */
  VOIDPTR(pMcDetectCard) = GetLibraryEntry("mcman", version, 5);   
  
  if (pMcDetectCard == NULL)
  {
   if (verbose) printf("mcdump: Unable to find the card detection routine.\n");
   return NO_RESIDENT_END;
  }
 }

 if (version < 0x200) printf("mcdump: Using old MCMAN module.\n");

 /* getting MCMAN's ___McGetCardType */
 VOIDPTR(pMcGetCardType) = GetLibraryEntry("mcman", version, 39); 
 if (pMcGetCardType == NULL)
 {
  if (verbose) printf("mcdump: Unable to get the MCMAN library entry.\n");
  return NO_RESIDENT_END;
 }

 LockMcman();

 /* detecting a memory card */ 
 r = pMcDetectCard(port, slot);
 if (verbose > 1) printf("mcdump: ___McDetectCard(%d, %d) result: %d.\n", port, slot, r);
 
 /* handling non-fatal errors */
 if ((r >= sceMcResNoFormat) && (r != sceMcResSucceed))
 {
  /* trying to detect a memory card again */
  r = pMcDetectCard(port, slot);
  if (verbose > 1) printf("mcdump: ___McDetectCard(%d, %d) result: %d.\n", port, slot, r);
 }
 
 FreeMcman();
 
 /* fatal error */
 if (r != sceMcResSucceed)
 {
  if (verbose) printf("mcdump: Unable to detect a memory card in port %d.\n", port);
  return NO_RESIDENT_END;
 }
 
 /* checking the memory card type */
 r = pMcGetCardType(port, slot);
 cardtype = r;
 
 switch (r)
 {
  /* No Memory Card */
  case sceMcTypeNoCard:
   mcbuffer = "No Memory Card";
   r = -1;
   break;
  /* Memory Card for PlayStation */
  case sceMcTypePS1:
   mcbuffer = "Memory Card for PlayStation";
   pMcGetSpec = McGetDefSpec1;
   r = 29;
   break;
  /* Memory Card for PlayStation 2 */
  case sceMcTypePS2:
   mcbuffer = "Memory Card for PlayStation 2";
   r = 18; /* entry number for MCMAN's ___McReadPage */
   break;
  /* PocketStation (PDA) */
  case sceMcTypePDA:
   mcbuffer = "PocketStation (PDA)";
   pMcGetSpec = McGetDefSpec1;
   r = 29;
   break;
  /* Unknown Memory Cards */
  default:
   mcbuffer = "Unknown Memory Card";
   r = -1;
   break;
 } 

 if (verbose) printf("mcdump: %s in port %d slot %d.\n", mcbuffer, port, slot);
 
 if (r < 0)
 {
  printf("mcdump: Unsupported memory card.\n");
  return NO_RESIDENT_END;
 }

 /* getting MCMAN's routine to read card page */
 VOIDPTR(pMcReadPage) = GetLibraryEntry("mcman", version, r);
 if (pMcReadPage == NULL)
 {
  if (verbose) printf("mcdump: Unable to get MCMAN libary entry to read card page.\n");
  return NO_RESIDENT_END;
 }

 pagesize = blocksize = cardsize = cardflags = 0;

 LockMcman();
 r = pMcGetSpec(port, 0, &pagesize, &blocksize, &cardsize, &cardflags);
 FreeMcman();
 if (verbose > 1) printf("mcdump: ___McGetSpec(%d, %d, ...) result: %d.\n", port, slot, r);
 
 if (r != sceMcResSucceed)
 {
  if (verbose) printf("mcdump: Unable to get memory card spec.\n");
  return NO_RESIDENT_END;
 }
  
 if (verbose) printf("mcdump: Memory Card %d,%d Spec:\n   Page Size: 0x%X\n   Block Size: 0x%X\n   Card Size: 0x%X\n   Card Flags: 0x%X\n", port, slot, pagesize, blocksize, cardsize, cardflags);

 blocks = cardsize / blocksize;
 erasebyte = (cardflags & 0x10) ? 0x0 : 0xFF;

 /* allocating memory for I/O buffer */
 mcbuffer = (char*)SysAlloc((pagesize + 0xFF) & ~(u_int)0xFF);
 if (mcbuffer == NULL)
 {
  if (verbose) printf("mcdump: Unable to allocate memory.\n");
  return NO_RESIDENT_END;
 }

 /* creating a card information file */
 if (info != NULL) 
 {
  register CardInfo *mcih;
  
  switch (cardtype)
  {
   case sceMcTypePS1:
   case sceMcTypePDA:
    r = PS1_CARDINFO_SIGN;
    break;
   case sceMcTypePS2:
    r = PS2_CARDINFO_SIGN;
    break;
   default:
    r = 0;
    break;
  }

  mcih = (CardInfo*)mcbuffer;
  
  mcih->sign = r;
  mcih->spec.PageSize = pagesize;
  mcih->spec.BlockSize = blocksize;
  mcih->spec.CardSize = cardsize;
  mcih->flag = cardflags;

  if (verbose) printf("mcdump: Creating info file: %s\n", info);

  fd = open(info, O_CREAT | O_RDWR | O_TRUNC, 0666);
  if (fd < 0)
  {
   SysFree(mcbuffer);
   
   if (verbose)
   {
    printf("mcdump: Unable to create file '%s'.\n", info);
    if (verbose > 1) printf("mcdump: I/O error: %d\n", fd);   
   }
   
   return NO_RESIDENT_END;
  }

  /* writing memory card information to file */
  r = write(fd, mcih, sizeof(CardInfo));
  if (r != sizeof(CardInfo))
  {
   close(fd);
   SysFree(mcbuffer);
   
   if (verbose)
   {
    printf("mcdump: Error writing card information file.\n");
    if ((verbose > 1) && (r < 0)) printf("mcdump: I/O error: %d\n", r);
   }

   return NO_RESIDENT_END;
  }
  
  close(fd);
 }
 
 /* creating an output file */ 
 fd = open(file, O_CREAT | O_RDWR | O_TRUNC, 0666);
 if (fd < 0)
 {
  SysFree(mcbuffer);
  
  if (verbose)
  {
   printf("mcdump: Unable to create file '%s'.\n", file);
   if (verbose > 1) printf("mcdump: I/O error: %d\n", fd);
  }
  
  return NO_RESIDENT_END;
 }

 if (verbose)
 {
  register int mcsize;
  register char *unit; 

  /* calculating Memory Card size in kilobytes */  
  mcsize = cardsize * pagesize / 1024;
  
  if (mcsize > 1024)
  {
   /* recalculating Memory Card size in megabytes */
   mcsize /= 1024;
   unit = "MB";
  }
  else unit = "KB";
 
  printf("mcdump: Dumping %d%s Memory Card in port %d slot %d.\nDo not remove the memory card or turn-off console.\n", mcsize, unit, port, slot);
 }
 
 LockMcman();
 
 /* copying a memory card content into the file */ 
 for (i = 0, page = 0; i < blocks; i ++)
 {
  register int e;

  if (verbose > 2) printf("mcdump: Reading page 0x%X\n", page);
 
  for (j = 0, e = 0; j < blocksize; j ++, page ++)
  {  
   if (!e)
   {
    /* reading memory card page */
    r = pMcReadPage(port, slot, page, mcbuffer);
    if (r != sceMcResSucceed)
    {
     if (verbose)
     {
      printf("mcdump: Memory Card read error at page 0x%X\n", page);
      if (verbose > 1) printf("mcdump: Memory Card read error code: %d\n", r);
     }
    
     /* setting the read error flag */
     if (j == 0) e = 1;
     /* creating erase-data */
     memset(mcbuffer, erasebyte, pagesize);
    }
   } 

   /* storing a page-data to the file */
   r = write(fd, mcbuffer, pagesize);
   if (r < 0)
   {
    if (verbose)
    {
     printf("mcdump: Unable to write output file.\n");
     if (verbose > 1) printf("mcdump: I/O error: %d\n", r);
    }
    
    break;
   }
  }    

  if (r < 0) break;
 }

 FreeMcman();
  
 close(fd);
 SysFree(mcbuffer);

 if (verbose) printf("mcdump: Memory Card dump complete.\n");

 return NO_RESIDENT_END;
}

void *GetLibraryEntry(char *libname, int version, int entryno)
{
 LibInfo lib;
 register void **exp;
 register int i;

 if (libname == NULL) return NULL;

 memset(&lib, 0, sizeof(LibInfo));
 lib.version = version & 0xFFFF;
 
 for (i = 0; (i < 8) && (libname[i]); i ++) lib.name[i] = libname[i];

 exp = (void **)QueryLibraryEntryTable(&lib);
 if (exp == NULL) return NULL;

 for (i = 0; exp[i]; i ++) {;}

 return (entryno < i) ? exp[entryno] : NULL;
}

/* getting default PlayStation Memory Card spec */
int McGetDefSpec1(int port, int slot, u_short *pagesize, u_short *blocksize, u_int *cardsize, u_char *flags)
{
 if (verbose) printf("mcdump: Using default PS1/PDA Memory Card 128KB spec for port %d slot %d.\n", port, slot);

 *pagesize = 0x80; /* Default page size: 128 bytes */
 *blocksize = 0x80; /* Default number of pages in block: 128 pages */
 *cardsize = 0x400; /* Default total number of pages: 1024 */
 *flags = 0x0; /* No flags for PS1/PDA cards */
 
 return sceMcResSucceed;
}

/* getting default PlayStation 2 Memory Card spec */
int McGetDefSpec2(int port, int slot, u_short *pagesize, u_short *blocksize, u_int *cardsize, u_char *flags)
{
 if (verbose) printf("mcdump: Using default PS2 Memory Card 8MB spec for port %d slot %d.\n", port, slot);
 
 *pagesize = 0x200; /* Default page size: 512 bytes */
 *blocksize = 0x10; /* Default number of pages in block: 16 pages */
 *cardsize = 0x4000; /* Default total number of pages: 16384 */
 *flags = 0x2B; /* Default memory card flags */
 
 return sceMcResSucceed;
}

void *SysAlloc(u_long size)
{
 int oldstate;
 register void *p;
 
 CpuSuspendIntr(&oldstate);
 p = AllocSysMemory(SMEM_Low, size, NULL);
 CpuResumeIntr(oldstate);

 return p;
}

int SysFree(void *area)
{
 int oldstate;
 register int r;

 CpuSuspendIntr(&oldstate);
 r = FreeSysMemory(area);
 CpuResumeIntr(oldstate);

 return r;
}

/* locking MCMAN's internal mutex when possible */
void LockMcman()
{
 if (pMcGetIoSema != NULL) WaitSema(pMcGetIoSema());
}

/* freeing MCMAN's internal mutex when possible */
void FreeMcman()
{
 if (pMcGetIoSema != NULL) SignalSema(pMcGetIoSema());
}
